#define addOccurencesForRelator     AddOfR
#define addRelatorSortedFromWord    AddRSW
#define addStrongGenerator          AddSGn
#define adjoinGenInverses           AdjGIv
#define adjoinInvImage              AdjIIm
#define adjoinInverseGen            AdjIGn
#define allocBooleanArrayBaseSize   AlcBAB
#define allocBooleanArrayDegree     AlcBAG
#define allocFactoredInt            AlcFI
#define allocField                  AlcFld
#define allocIntArrayBaseSize       AlcIAB
#define allocIntArrayDegree         AlcIAD
#define allocLongArrayBaseSize      AlcLAB
#define allocOccurenceOfGen         AlcOoG
#define allocPartition              AlcPar
#define allocPartitionStack         AlcPaS
#define allocPermGroup              AlcPG
#define allocPermutation            AlcPer
#define allocPointSet               AlcPS
#define allocPtrArrayBaseSize       AlcPAB
#define allocPtrArrayDegree         AlcPAD
#define allocPtrArrayWordSize       AlcPAW
#define allocRBase                  AlcRB
#define allocRelator                AlcRel
#define allocRPriorityQueue         AlcRPQ
#define allocRefinementArrayDegree  AlcRAD
#define allocWord                   AlcWor
#define augmentedMatrix             AugMat
#define assignGenName               AsnGN
#define bitSetAt                    BitStA
#define bitSetBelow                 BitStB
#define buildField                  BldFld
#define cellNumberAtDepth           CelNAD
#define centralizer                 centrl
#define changeBase                  ChaBas
#define checkCompileOptions         CkCpOp
#define checkConjugacy              ChkCjg
#define checkConjugacyInGroup       ChkCGp
#define chooseNextBasePoint         CNxBPt
#define codeContainsVector          CodCVe
#define commutatorGroup             ComGrp
#define compressAtLevel             CmprAL
#define compressGroup               CmprGp
#define computeCosetRep             CpCRep
#define computeSchreierGen          CpShGn
#define computeSubgroup             CpSG
#define conjugateGroupByPerm        CjGrpP
#define conjugatePermByGroup        CjPerG
#define conjugatePermByPerm         CjPerP
#define conjugatingElement          CnjElt
#define constructAllOrbitInfo       CstAOI
#define constructBasicOrbit         CstBO
#define constructOrbitPartition     CstOP
#define constructRBase              CstRB
#define copyEssential               CopEss
#define copyOfPermGroup             CopOPG
#define copyOfPermutation           CopOP
#define copyPermutation             CopPer
#define currentExtraCosets          CuExCs
#define deletePartition             DelPar
#define deletePartitionStack        DelPS
#define deletePermGroup             DelPG
#define deletePermutation           DelPer
#define deleteRBase                 DelRB
#define deleteRPriorityQueue        DelRPQ
#define depthGreaterThan            DepGt
#define designAutoGroup             DesAGp
#define designIsomorphism           DesIso
#define equivCoset                  EqvCos
#define errorMessage                ErrMes
#define errorMessage1i              ErrM1i
#define errorMessage1s              ErrM1s
#define essentialAboveLevel         EssAbL
#define essentialAtLevel            EssAtL
#define essentialBelowLevel         EssBeL
#define expandGenerators            ExpGen
#define expandSGS                   ExpSGS
#define extendBasicOrbit            ExtBO
#define extraCosetsAtLevel          ExCsLv
#define factDivide                  FacDiv
#define factEqual                   FacEql
#define factMultiply                FacMul
#define factorize                   Factiz
#define findConsequences            FnCnsq
#define fixesBasicOrbit             FixBO
#define forceCollapse               FCalps
#define freeBooleanArrayBaseSize    FreBAB
#define freeBooleanArrayDegree      FreBAD
#define freeCosHeader               FreCsH
#define freeFactoredInt             FreFI
#define freeField                   FreFld
#define freeIntArrayBaseSize        FreIAB
#define freeIntArrayDegree          FreIAD
#define freeLongArrayBaseSize       FreLAB
#define freeOccurenceOfGen          FreOoG
#define freePartition               FrePar
#define freePartitionStack          FrePaS
#define freePermGroup               FrePG
#define freePermutation             FrePer
#define freePointSet                FrePS
#define freePtrArrayBaseSize        FrePAB
#define freePtrArrayDegree          FrePAD
#define freePtrArrayWordSize        FrePAW
#define freeRBase                   FreRB
#define freeRelator                 FreRel
#define freeRPriorityQueue          FreRPQ
#define freeRefinementArrayDegree   FreRAD
#define freeWord                    FreWrd
#define genCount                    GenCnt
#define genExpandingBasicOrbit      GenEBO
#define groupCentralizer            GpCent
#define informCosetRep              InfCR
#define informGroup                 InfGrp
#define informNewGenerator          InfNG
#define informNewRelator            InfNR
#define informOptions               InfOpt
#define informRBase                 InfRB
#define informSTCSSummary           InfSTC
#define informStatistics            InfStt
#define informSubgroup              InfSG
#define informTime                  IntTim
#define initFromPartnStack          IniPaS
#define initializeBase              IniBas 
#define initializePartnStabRefn     IniPSR
#define initializeOrbRefine         IniOR
#define initializeSeed              IniSed
#define initializeSetStabRefn       IniSSR
#define initializeStorageManager    IniSM
#define initializeUPartnStabRefine  IniUPS
#define insertBasePoint             InsBP
#define intersection                Inters
#define isBaseImage                 IsBImg
#define isCentralizedBy             IsCenB
#define isCodeIsomorphism           IsCIso
#define isDoublyTransitive          IsDbTr
#define isElementOf                 IsEltO
#define isFixedPointOf              IsFxP
#define isIdentity                  IsIden
#define isIdentityElt               IsIElt
#define isInvolution                IsInvo
#define isInvolutoryElt             IsIvEl
#define isMatrix01Isomorphism       IsMIso
#define isNontrivialGroup           IsNtGp
#define isNormalizedBy              IsNorB
#define isOrbReducible              IsOrRd
#define isPartnStabReducible        IsPaRd
#define isPointStabReducible        IsPSRd
#define isSetStabReducible          IsSSRd
#define isSubgroupOf                IsSGrp
#define isValidName                 IsValN
#define isValidPermutation          IsValP
#define is_sSSPointReducible        IsSPRd
#define is_sSSReducible             IsSRd
#define leftMultiply                LftMul
#define levelIn                     LevIn
#define linkEssentialGens           LnkEG
#define linkEssentialGensAtLevel    LnkEGL
#define linkGensAtLevel             LnkGAL
#define lowerCase                   LwrCas
#define makeDefinition              MakDef
#define makeEmpty                   MakEmp
#define makeEssentialAtLevel        MEsAtL
#define makeNotEssentialAll         MEsAll
#define makeNotEssentialAtAboveLevel MEsAAL
#define makeNotEssentialAtLevel     MNEAtL
#define makeUnknownEssential        MUnkEs
#define matrixAutoGroup             MatAut
#define matrixIsomorphism           MatIso
#define meanCosetRepLen             MenCRL
#define minimalPointOfOrbit         MinPOO
#define newCellPartitionStack       NewCPS
#define newIdentityPerm             NewIP
#define newPartitionStack           NewPaS
#define newRBase                    NewRB
#define newRelatorFromWord          NewRfW
#define newRPriorityQueue           NewRPQ
#define newTrivialPermGroup         NewTPG
#define newTrivialWord              NewTW
#define newUndefinedPerm            NewUP
#define newZeroMatrix               NewZMa
#define nkReadToken                 NkRedT
#define normalClosure               NorCls
#define numberOfCells               NoOCel
#define onFreeList                  OnFreL
#define orbRefine                   OrbRef
#define parseLibraryName            ParsLN  
#define partnStabilizer             ParStb
#define partnStabRefine             PaStbR
#define partnImage                  ParImg
#define permMapping                 PerMap
#define permOrder                   PerOrd
#define pointMovedBy                PtMovB
#define pointStabRefine             PtStbR
#define popToHeight                 PopTHt
#define primeList                   PrmLst
#define processCoincidence          PrCoin
#define prodOrderBounded            PrOrdB
#define ptStabFamily                PtStFm
#define raisePermToPower            RasPTP
#define randGroupPerm               RandGP
#define randGroupWord               RandGW
#define randInteger                 RandIn
#define randomSchreier              RandSh
#define read01Matrix                Red01M
#define readCode                    RedCod
#define readCyclePerm               RedCyP
#define readDesign                  RedDes
#define readFactoredInt             RedFI
#define readImagePerm               RedIP
#define readPartition               RedPar
#define readPerm                    RedPm
#define readPermGroup               RedPG
#define readPermutation             RedPer
#define readPointList               RedPL
#define readPointSet                RedPS
#define readToken                   RedTok
#define reconstructBasicOrbit       RecBO
#define reduceBasis                 RedBas
#define reduceWrtGroup              RedWGp
#define relatorLevel                RelLev
#define relatorSelection            RelSel
#define removeIdentityGens          RmvIGn
#define removeMin                   RmvMin
#define removeRedunSGens            RmvRSG
#define replaceByPower              RepPwr
#define resetTable                  ResetT
#define restrictBasePoints          ResBPt
#define rightMultiply               RgtMul
#define rightMultiplyInv            RgtMIv
#define schreierToddCoxeterSims     SchTCS
#define shiftPriority               SftPri
#define shiftSelection              SftSel
#define sReadToken                  SRdTok
#define sSSPoint                    SSSPnt
#define sSS_Partn                   SSSPar
#define sUnreadToken                SUrTok
#define setImage                    SetImg
#define setInputFile                SetInF
#define setInputString              SetInS
#define setOutputFile               SetOtF
#define setStabilizer               SetStb
#define showLimits                  ShoLim
#define symmetricLength             SymLen
#define symmetricWordLength         SymWLn
#define traceNewRelator             TrNRel
#define unreadToken                 URdTok
#define uPartnImage                 UPrImg 
#define uPartnStabilizer            UprStb
#define verifyCosetList             VerCLs
#define write01Matrix               Wrt01M
#define writeCode                   WrtCod
#define writeCyclePerm              WrtCyP
#define writeDesign                 WrtDes
#define writeFactoredInt            WrtFI
#define writeImagePerm              WrtIP
#define writeImageMonomialPerm      WrtIMP
#define writePartition              WrtPar
#define writePermGroup              WrtPG
#define writePermGroupRestricted    WrtPGR
#define writePermutation            WrtPer
#define writePermutationRestricted  WrtPRs
#define writePointSet               WrtPS
#define xComputeSchreierGen         XCpShG
#define xFindConsequences           XFndCs
#define xPopToLevel                 XPopLv
#define xTraceNewRelator            XTrNRl
