#ifndef COPY
#define COPY

extern Permutation *copyOfPermutation(
   const Permutation *const oldPerm)   /* The permutation to be copied. */
;

extern PermGroup *copyOfPermGroup(
   const PermGroup *const oldGroup)    /* The group being copied. */
;

#endif
